"""Tests for brother package."""
import json
from datetime import UTC, datetime
from unittest.mock import Mock, patch

import pytest
from pysnmp.smi.rfc1902 import ObjectType

from brother import OIDS, Brother, SnmpError, UnsupportedModelError

HOST = "localhost"
INVALID_HOST = "foo.local"
TEST_TIME = datetime(2019, 11, 11, 9, 10, 32, tzinfo=UTC)


@pytest.mark.asyncio
async def test_hl_l2340dw_model() -> None:
    """Test with valid data from HL-L2340DW printer with invalid printer_type."""
    with open("tests/fixtures/hl-l2340dw.json", encoding="utf-8") as file:
        data = json.load(file)

    with patch("brother.Brother._get_data", return_value=data) as mock_update, patch(
        "brother.datetime", now=Mock(return_value=TEST_TIME)
    ), patch("brother.Brother.initialize"):
        brother = await Brother.create(HOST, printer_type="foo")
        sensors = await brother.async_update()
        assert mock_update.call_count == 1

        # second update to test uptime logic
        sensors = await brother.async_update()
        assert mock_update.call_count == 2

    brother.shutdown()

    assert brother.model == "HL-L2340DW"
    assert brother.firmware == "1.17"
    assert brother.serial == "serial_number"
    assert sensors.status == "oczekiwanie"
    assert sensors.black_toner == 80
    assert sensors.page_counter == 986
    assert sensors.uptime.isoformat() == "2019-09-24T12:14:56+00:00"


@pytest.mark.asyncio
async def test_dcp_l3550cdw_model() -> None:
    """Test with valid data from DCP-L3550CDW printer."""
    with open("tests/fixtures/dcp-l3550cdw.json", encoding="utf-8") as file:
        data = json.load(file)
    brother = Brother(HOST)

    with patch("brother.Brother._get_data", return_value=data):
        sensors = await brother.async_update()

    brother.shutdown()

    assert brother.model == "DCP-L3550CDW"
    assert brother.firmware == "J1906051424"
    assert brother.serial == "serial_number"
    assert sensors.status == "mało toneru (y)"
    assert sensors.black_toner == 30
    assert sensors.yellow_toner == 10
    assert sensors.magenta_toner == 10
    assert sensors.cyan_toner == 10
    assert sensors.page_counter == 1611


@pytest.mark.asyncio
async def test_dcp_j132w_model() -> None:
    """Test with valid data from DCP-J132W printer."""
    with open("tests/fixtures/dcp-j132w.json", encoding="utf-8") as file:
        data = json.load(file)
    brother = Brother(HOST, printer_type="ink")

    with patch("brother.Brother._get_data", return_value=data):
        sensors = await brother.async_update()

    brother.shutdown()

    assert brother.model == "DCP-J132W"
    assert brother.firmware == "Q1906110144"
    assert brother.serial == "serial_number"
    assert sensors.status == "ready"
    assert sensors.black_ink == 80
    assert sensors.page_counter == 879


@pytest.mark.asyncio
async def test_mfc_5490cn_model() -> None:
    """Test with valid data from MFC-5490CN printer with no charset data."""
    with open("tests/fixtures/mfc-5490cn.json", encoding="utf-8") as file:
        data = json.load(file)
    brother = Brother(HOST, printer_type="ink")
    brother._legacy = True  # pylint:disable=protected-access

    with patch("brother.Brother._get_data", return_value=data), patch(
        "brother.datetime", now=Mock(return_value=TEST_TIME)
    ):
        sensors = await brother.async_update()

    brother.shutdown()

    assert brother.model == "MFC-5490CN"
    assert brother.firmware == "U1005271959VER.E"
    assert brother.serial == "serial_number"
    assert sensors.status == "sleep mode"
    assert sensors.page_counter == 8989
    assert sensors.uptime.isoformat() == "2019-11-02T23:44:02+00:00"


@pytest.mark.asyncio
async def test_dcp_l2540dw_model() -> None:
    """Test with valid data from DCP-L2540DN printer with status in Russian."""
    with open("tests/fixtures/dcp-l2540dn.json", encoding="utf-8") as file:
        data = json.load(file)
    brother = Brother(HOST, printer_type="laser")

    with patch("brother.Brother._get_data", return_value=data):
        sensors = await brother.async_update()

    brother.shutdown()

    assert brother.model == "DCP-L2540DN"
    assert brother.firmware == "R1906110243"
    assert brother.serial == "serial_number"
    assert sensors.status == "спящий режим"
    assert sensors.black_toner_remaining == 55
    assert sensors.page_counter == 333


@pytest.mark.asyncio
async def test_dcp_7070dw_model() -> None:
    """Test with valid data from DCP-7070DW printer with status in Dutch."""
    with open("tests/fixtures/dcp-7070dw.json", encoding="utf-8") as file:
        data = json.load(file)
    brother = Brother(HOST, printer_type="laser")

    with patch("brother.Brother._get_data", return_value=data), patch(
        "brother.datetime", now=Mock(return_value=TEST_TIME)
    ):
        sensors = await brother.async_update()

    assert brother.model == "DCP-7070DW"
    assert brother.firmware == "U1307022128VER.J"
    assert brother.serial == "serial_number"
    assert sensors.status == "stap. kopieën:01"
    assert sensors.black_toner_remaining == 72
    assert sensors.page_counter == 2652
    assert sensors.drum_counter == 1603
    assert sensors.drum_remaining_life == 88
    assert sensors.drum_remaining_pages == 10397
    assert sensors.uptime.isoformat() == "2018-11-30T13:43:26+00:00"

    # test uptime logic, uptime increased by 10 minutes
    data["1.3.6.1.2.1.1.3.0"] = "2987742561"
    with patch("brother.Brother._get_data", return_value=data), patch(
        "brother.datetime", now=Mock(return_value=TEST_TIME)
    ):
        sensors = await brother.async_update()

    brother.shutdown()

    assert sensors.uptime.isoformat() == "2018-11-30T13:53:26+00:00"


@pytest.mark.asyncio
async def test_mfc_j680dw_model() -> None:
    """Test with valid data from MFC-J680DW printer with status in Turkish."""
    with open("tests/fixtures/mfc-j680dw.json", encoding="utf-8") as file:
        data = json.load(file)
    brother = Brother(HOST, printer_type="ink")

    with patch("brother.Brother._get_data", return_value=data):
        sensors = await brother.async_update()

    brother.shutdown()

    assert brother.model == "MFC-J680DW"
    assert brother.firmware == "U1804191714VER.J"
    assert brother.serial == "serial_number"
    assert sensors.status == "uyku"
    assert sensors.black_ink == 47
    assert sensors.color_counter == 491


@pytest.mark.asyncio
async def test_dcp_9020cdw_model() -> None:
    """Test with valid data from DCP-9020CDW printer."""
    with open("tests/fixtures/dcp-9020cdw.json", encoding="utf-8") as file:
        data = json.load(file)
    brother = Brother(HOST, printer_type="laser")

    with patch("brother.Brother._get_data", return_value=data):
        sensors = await brother.async_update()

    brother.shutdown()

    assert brother.model == "DCP-9020CDW"
    assert brother.firmware == "ZA1811191217"
    assert brother.serial == "E71833C4J372261"
    assert sensors.status == "tryb uśpienia"
    assert sensors.cyan_drum_remaining_life == 68
    assert sensors.cyan_drum_counter == 4939
    assert sensors.cyan_drum_remaining_pages == 10061


@pytest.mark.asyncio
async def test_hl_2270dw_model() -> None:
    """Test with valid data from HL-2270DW printer."""
    with open("tests/fixtures/hl-2270dw.json", encoding="utf-8") as file:
        data = json.load(file)
    brother = Brother(HOST, printer_type="laser")

    with patch("brother.Brother._get_data", return_value=data):
        sensors = await brother.async_update()

    brother.shutdown()

    assert brother.model == "HL-2270DW"
    assert brother.firmware == "1.16"
    assert brother.serial == "serial_number"
    assert sensors.status == "sleep"
    assert sensors.page_counter == 4191
    assert sensors.drum_remaining_pages == 7809


@pytest.mark.asyncio
async def test_mfc_t910dw_model() -> None:
    """Test with valid data from MFC-T910DW printer."""
    with open("tests/fixtures/mfc-t910dw.json", encoding="utf-8") as file:
        data = json.load(file)
    brother = Brother(HOST, printer_type="ink")

    with patch("brother.Brother._get_data", return_value=data):
        sensors = await brother.async_update()

    brother.shutdown()

    assert brother.model == "MFC-T910DW"
    assert brother.firmware == "M2009041848"
    assert brother.serial == "serial_number"
    assert sensors.status == "oczekiwanie"
    assert sensors.page_counter == 3384
    assert sensors.color_counter == 3199
    assert sensors.bw_counter == 185
    assert sensors.duplex_unit_pages_counter == 1445
    assert sensors.black_ink_status == 1
    assert sensors.cyan_ink_status == 1
    assert sensors.magenta_ink_status == 1
    assert sensors.yellow_ink_status == 1


@pytest.mark.asyncio
async def test_hl_5350dn_model() -> None:
    """Test with valid data from HL-5350DN printer."""
    with open("tests/fixtures/hl-5350dn.json", encoding="utf-8") as file:
        data = json.load(file)
    brother = Brother(HOST, printer_type="laser")

    with patch("brother.Brother._get_data", return_value=data):
        sensors = await brother.async_update()

    brother.shutdown()

    assert brother.model == "HL-5350DN"
    assert brother.firmware is None
    assert brother.serial == "serial_number"
    assert sensors.status == "energiesparen trommel ersetz."
    assert sensors.page_counter == 69411


@pytest.mark.asyncio
async def test_invalid_data() -> None:
    """Test with invalid data from printer."""
    with open("tests/fixtures/invalid.json", encoding="utf-8") as file:
        data = json.load(file)
    brother = Brother(HOST)

    with patch("brother.Brother._get_data", return_value=data), pytest.raises(
        UnsupportedModelError
    ):
        await brother.async_update()

    brother.shutdown()


@pytest.mark.asyncio
async def test_incomplete_data() -> None:
    """Test with incomplete data from printer."""
    with open("tests/fixtures/incomplete.json", encoding="utf-8") as file:
        data = json.load(file)
    brother = Brother(HOST)

    with patch("brother.Brother._get_data", return_value=data):
        await brother.async_update()

    brother.shutdown()


@pytest.mark.asyncio
async def test_empty_data() -> None:
    """Test with empty data from printer."""
    brother = Brother(HOST)

    with patch("brother.Brother._get_data", return_value=None), pytest.raises(
        SnmpError
    ):
        await brother.async_update()

    brother.shutdown()


@pytest.mark.asyncio
async def test_invalid_host() -> None:
    """Test with invalid host."""
    with patch(
        "brother.Brother.initialize", side_effect=ConnectionError("Connection Error")
    ), pytest.raises(ConnectionError):
        await Brother.create(INVALID_HOST)


@pytest.mark.asyncio
async def test_snmp_error() -> None:
    """Test with raise SnmpError."""
    with patch(
        "brother.Brother.initialize", side_effect=SnmpError("SNMP Error")
    ), pytest.raises(SnmpError):
        await Brother.create(HOST)


@pytest.mark.asyncio
async def test_unsupported_model() -> None:
    """Test with unsupported printer model."""
    with pytest.raises(UnsupportedModelError):
        Brother(HOST, model="mfc-8660dn")


def test_iterate_oids() -> None:
    """Test iterate_oids function."""
    brother = Brother(HOST)
    oids = OIDS.values()
    result = list(brother._iterate_oids(oids))  # pylint:disable=protected-access

    assert len(result) == 10
    for item in result:
        assert isinstance(item, ObjectType)



"""Python wrapper for getting data from Brother laser and inkjet printers via SNMP."""
import logging
import re
from collections.abc import Generator, Iterable
from contextlib import suppress
from datetime import UTC, datetime, timedelta
from typing import TYPE_CHECKING, Any, Self, cast

import pysnmp.hlapi.asyncio as hlapi
from dacite import from_dict
from pysnmp.error import PySnmpError
from pysnmp.hlapi.asyncio.cmdgen import lcd
from pysnmp.smi.rfc1902 import ObjectType

from .const import (
    ATTR_CHARSET,
    ATTR_COUNTERS,
    ATTR_FIRMWARE,
    ATTR_MAINTENANCE,
    ATTR_MODEL,
    ATTR_NEXTCARE,
    ATTR_PAGE_COUNT,
    ATTR_SERIAL,
    ATTR_STATUS,
    ATTR_UPTIME,
    CHARSET_MAP,
    OIDS,
    OIDS_HEX,
    PERCENT_VALUES,
    PRINTER_TYPES,
    UNSUPPORTED_MODELS,
    UPTIME_FLUCTUATION_SECONDS,
    VALUES_COUNTERS,
    VALUES_INK_MAINTENANCE,
    VALUES_LASER_MAINTENANCE,
    VALUES_LASER_NEXTCARE,
)
from .exceptions import SnmpError, UnsupportedModelError
from .model import BrotherSensors

_LOGGER = logging.getLogger(__name__)

REGEX_MODEL_PATTERN = re.compile(r"MDL:(?P<model>[\w\-]+)")
CHUNK_SIZE = 14
LEGACY_CHUNK_SIZE = 10


class Brother:
    """Main class to perform snmp requests to printer."""

    def __init__(
        self,
        host: str,
        port: int = 161,
        printer_type: str = "laser",
        model: str | None = None,
        snmp_engine: hlapi.SnmpEngine = None,
    ) -> None:
        """Initialize."""
        if model:
            _LOGGER.debug("Model: %s", model)
            for unsupported_model in UNSUPPORTED_MODELS:
                if unsupported_model in model.lower():
                    raise UnsupportedModelError(
                        "It seems that this printer model is not supported"
                    )

        if printer_type not in PRINTER_TYPES:
            _LOGGER.warning("Wrong printer_type argument, 'laser' was used")
            self._printer_type = "laser"
        else:
            self._printer_type = printer_type

        self._legacy = False

        self.firmware: str | None = None
        self.model: str | None = None
        self.serial: str
        self._host = host
        self._port = port
        self._last_uptime: datetime | None = None
        self._snmp_engine = snmp_engine
        self._oids: list[ObjectType] = []

    @classmethod
    async def create(
        cls,
        host: str,
        port: int = 161,
        printer_type: str = "laser",
        model: str | None = None,
        snmp_engine: hlapi.SnmpEngine = None,
    ) -> Self:
        """Create a new device instance."""
        instance = cls(host, port, printer_type, model, snmp_engine)
        await instance.initialize()
        return instance

    async def initialize(self) -> None:
        """Initialize snmp_engine and check which OIDs are supported."""
        _LOGGER.debug("Initializing device %s", self._host)

        if not self._snmp_engine:
            self._snmp_engine = hlapi.SnmpEngine()

        oids = list(self._iterate_oids(OIDS.values()))

        try:
            request_args = [
                self._snmp_engine,
                hlapi.CommunityData("public", mpModel=0),
                hlapi.UdpTransportTarget(
                    (self._host, self._port), timeout=2, retries=10
                ),
                hlapi.ContextData(),
            ]
        except PySnmpError as err:
            raise ConnectionError(err) from err

        while True:
            get_result = await hlapi.getCmd(*request_args, *oids)
            _, errstatus, errindex, _ = get_result

            if str(errstatus) == "noSuchName":
                # 5 and 8 are indexes from OIDS consts, model and serial are obligatory
                if errindex in (5, 8):
                    raise UnsupportedModelError(
                        "It seems that this printer model is not supported"
                    )

                oids.pop(errindex - 1)
                continue

            break

        self._oids = oids

    async def async_update(self) -> BrotherSensors:
        """Update data from printer."""
        if not (raw_data := await self._get_data()):
            raise SnmpError("The printer did not return data")

        _LOGGER.debug("RAW data: %s", raw_data)

        data: dict[str, str | int | datetime | None] = {}

        try:
            model_match = re.search(REGEX_MODEL_PATTERN, raw_data[OIDS[ATTR_MODEL]])
            if TYPE_CHECKING:
                assert model_match is not None
            self.model = data[ATTR_MODEL] = cast(str, model_match.group("model"))
            self.serial = data[ATTR_SERIAL] = raw_data[OIDS[ATTR_SERIAL]]
        except (TypeError, AttributeError, AssertionError) as err:
            raise UnsupportedModelError(
                "It seems that this printer model is not supported"
            ) from err

        self.firmware = data[ATTR_FIRMWARE] = raw_data.get(OIDS[ATTR_FIRMWARE])

        charset = CHARSET_MAP.get(raw_data.get(OIDS[ATTR_CHARSET], "unknown"), "roman8")

        if status := raw_data[OIDS[ATTR_STATUS]]:
            data[ATTR_STATUS] = (
                self._cleanse_status(status).encode("latin1").decode(charset).lower()
            )

        try:
            uptime = int(cast(str, raw_data.get(OIDS[ATTR_UPTIME]))) / 100
        except TypeError:
            pass
        else:
            if self._last_uptime:
                new_uptime = (datetime.now(tz=UTC) - timedelta(seconds=uptime)).replace(
                    microsecond=0, tzinfo=UTC
                )
                if (
                    abs((new_uptime - self._last_uptime).total_seconds())
                    > UPTIME_FLUCTUATION_SECONDS
                ):
                    data[ATTR_UPTIME] = self._last_uptime = new_uptime
                else:
                    data[ATTR_UPTIME] = self._last_uptime
            else:
                data[ATTR_UPTIME] = self._last_uptime = (
                    datetime.now(tz=UTC) - timedelta(seconds=uptime)
                ).replace(microsecond=0, tzinfo=UTC)
        if self._legacy:
            if self._printer_type == "laser":
                data.update(
                    dict(
                        self._iterate_data_legacy(
                            raw_data.get(OIDS[ATTR_MAINTENANCE], {}),
                            VALUES_LASER_MAINTENANCE,
                        )
                    )
                )
            if self._printer_type == "ink":
                data.update(
                    dict(
                        self._iterate_data_legacy(
                            raw_data.get(OIDS[ATTR_MAINTENANCE], {}),
                            VALUES_INK_MAINTENANCE,
                        )
                    )
                )
        else:
            if self._printer_type == "laser":
                data.update(
                    dict(
                        self._iterate_data(
                            raw_data.get(OIDS[ATTR_COUNTERS], {}), VALUES_COUNTERS
                        )
                    )
                )
                data.update(
                    dict(
                        self._iterate_data(
                            raw_data.get(OIDS[ATTR_MAINTENANCE], {}),
                            VALUES_LASER_MAINTENANCE,
                        )
                    )
                )
                data.update(
                    dict(
                        self._iterate_data(
                            raw_data.get(OIDS[ATTR_NEXTCARE], {}), VALUES_LASER_NEXTCARE
                        )
                    )
                )
            if self._printer_type == "ink":
                data.update(
                    dict(
                        self._iterate_data(
                            raw_data.get(OIDS[ATTR_COUNTERS], {}), VALUES_COUNTERS
                        )
                    )
                )
                data.update(
                    dict(
                        self._iterate_data(
                            raw_data.get(OIDS[ATTR_MAINTENANCE], {}),
                            VALUES_INK_MAINTENANCE,
                        )
                    )
                )
        # page counter for old printer models
        with suppress(ValueError):
            if not data.get(ATTR_PAGE_COUNT) and raw_data.get(OIDS[ATTR_PAGE_COUNT]):
                data[ATTR_PAGE_COUNT] = int(
                    cast(str, raw_data.get(OIDS[ATTR_PAGE_COUNT]))
                )

        _LOGGER.debug("Data: %s", data)

        result: BrotherSensors = from_dict(BrotherSensors, data)

        return result

    def shutdown(self) -> None:
        """Unconfigure SNMP engine."""
        if self._snmp_engine:
            lcd.unconfigure(self._snmp_engine, None)

    async def _get_data(self) -> dict[str, Any]:
        """Retrieve data from printer."""
        raw_data: dict[str, str | list[str]] = {}

        try:
            request_args = [
                self._snmp_engine,
                hlapi.CommunityData("public", mpModel=0),
                hlapi.UdpTransportTarget(
                    (self._host, self._port), timeout=2, retries=10
                ),
                hlapi.ContextData(),
            ]
            get_result = await hlapi.getCmd(*request_args, *self._oids)
            errindication, errstatus, errindex, restable = get_result
        except PySnmpError as err:
            raise ConnectionError(err) from err
        if errindication:
            raise SnmpError(errindication)
        if errstatus:
            raise SnmpError(f"{errstatus}, {errindex}")
        for resrow in restable:
            if str(resrow[0]) in OIDS_HEX:
                # asOctets gives bytes data b'c\x01\x04\x00\x00\x00\x01\x11\x01\x04\x00\
                # x00\x05,A\x01\x04\x00\x00"\xc41\x01\x04\x00\x00\x00\x01o\x01\x04\x00\
                # x00\x19\x00\x81\x01\x04\x00\x00\x00F\x86\x01\x04\x00\x00\x00\n\xff'
                data = resrow[-1].asOctets()
                # convert to string without checksum FF at the end, gives
                # '630104000000011101040000052c410104000022c4310104000000016f01040000190
                #  0810104000000468601040000000a'
                data_str = "".join([f"{x:02x}" for x in data])[0:-2]
                # split to 14 digits words in list, gives ['63010400000001',
                # '1101040000052c', '410104000022c4', '31010400000001',
                # '6f010400001900', '81010400000046', '8601040000000a']
                result = [
                    data_str[ind : ind + CHUNK_SIZE]
                    for ind in range(0, len(data_str), CHUNK_SIZE)
                ]
                # map sensors names to OIDs
                raw_data[str(resrow[0])] = result
            else:
                raw_data[str(resrow[0])] = str(resrow[-1])
        # for legacy printers
        for resrow in restable:
            if str(resrow[0]) == OIDS[ATTR_MAINTENANCE]:
                # asOctets() gives bytes data
                data = resrow[-1].asOctets()
                # convert to string without checksum FF at the end, gives
                # 'a101020414a201020c14a301020614a401020b14'
                data_str = "".join([f"{x:02x}" for x in data])[0:-2]
                if self._legacy_printer(data_str):
                    self._legacy = True
                    # split to 10 digits words in list, gives ['a101020414',
                    # 'a201020c14', 'a301020614', 'a401020b14']
                    result = [
                        data_str[ind : ind + LEGACY_CHUNK_SIZE]
                        for ind in range(0, len(data_str), LEGACY_CHUNK_SIZE)
                    ]
                    # map sensors names to OIDs
                    raw_data[str(resrow[0])] = result
                    break
        return raw_data

    @staticmethod
    def _legacy_printer(string: str) -> bool:
        """Return True if printer is legacy."""
        length = len(string)
        nums = [x * 10 for x in range(length // 10)][1:]
        if results := [string[i - 2 : i] == "14" for i in nums]:
            return all(item for item in results)
        return False

    @staticmethod
    def _iterate_oids(oids: Iterable) -> Generator:
        """Iterate OIDS to retrieve from printer."""
        for oid in oids:
            yield hlapi.ObjectType(hlapi.ObjectIdentity(oid))

    @staticmethod
    def _iterate_data(iterable: Iterable, values_map: dict[str, str]) -> Generator:
        """Iterate data from hex words."""
        for item in iterable:
            # first byte means kind of sensor, last 4 bytes means value
            if item[:2] in values_map:
                if values_map[item[:2]] in PERCENT_VALUES:
                    yield (values_map[item[:2]], round(int(item[-8:], 16) / 100))
                else:
                    yield (values_map[item[:2]], int(item[-8:], 16))

    @staticmethod
    def _iterate_data_legacy(
        iterable: Iterable, values_map: dict[str, str]
    ) -> Generator:
        """Iterate data from hex words for legacy printers."""
        for item in iterable:
            # first byte means kind of sensor, last 4 bytes means value
            if item[:2] in values_map:
                yield (
                    values_map[item[:2]],
                    round(int(item[6:8], 16) / int(item[8:10], 16) * 100),
                )

    @staticmethod
    def _cleanse_status(status: str) -> str:
        """Cleanse and format status."""
        return " ".join(status.split()).strip()



"""Type definitions for Brother."""
from dataclasses import dataclass
from datetime import datetime


@dataclass(frozen=True)
class BrotherSensors:
    """Brother Sensors class."""

    model: str
    serial: str
    belt_unit_remaining_life: int | None = None
    belt_unit_remaining_pages: int | None = None
    black_counter: int | None = None
    black_drum_counter: int | None = None
    black_drum_remaining_life: int | None = None
    black_drum_remaining_pages: int | None = None
    black_ink_remaining: int | None = None
    black_ink_status: int | None = None
    black_ink: int | None = None
    black_toner_remaining: int | None = None
    black_toner_status: int | None = None
    black_toner: int | None = None
    bw_counter: int | None = None
    color_counter: int | None = None
    cyan_counter: int | None = None
    cyan_drum_counter: int | None = None
    cyan_drum_remaining_life: int | None = None
    cyan_drum_remaining_pages: int | None = None
    cyan_ink_remaining: int | None = None
    cyan_ink_status: int | None = None
    cyan_ink: int | None = None
    cyan_toner_remaining: int | None = None
    cyan_toner_status: int | None = None
    cyan_toner: int | None = None
    drum_counter: int | None = None
    drum_remaining_life: int | None = None
    drum_remaining_pages: int | None = None
    drum_status: int | None = None
    duplex_unit_pages_counter: int | None = None
    firmware: str | None = None
    fuser_remaining_life: int | None = None
    fuser_unit_remaining_pages: int | None = None
    image_counter: int | None = None
    laser_remaining_life: int | None = None
    laser_unit_remaining_pages: int | None = None
    magenta_counter: int | None = None
    magenta_drum_counter: int | None = None
    magenta_drum_remaining_life: int | None = None
    magenta_drum_remaining_pages: int | None = None
    magenta_ink_remaining: int | None = None
    magenta_ink_status: int | None = None
    magenta_ink: int | None = None
    magenta_toner_remaining: int | None = None
    magenta_toner_status: int | None = None
    magenta_toner: int | None = None
    page_counter: int | None = None
    pf_kit_1_remaining_life: int | None = None
    pf_kit_1_remaining_pages: int | None = None
    pf_kit_mp_remaining_life: int | None = None
    pf_kit_mp_remaining_pages: int | None = None
    status: str | None = None
    uptime: datetime | None = None
    yellow_counter: int | None = None
    yellow_drum_counter: int | None = None
    yellow_drum_remaining_life: int | None = None
    yellow_drum_remaining_pages: int | None = None
    yellow_ink_remaining: int | None = None
    yellow_ink_status: int | None = None
    yellow_ink: int | None = None
    yellow_toner_remaining: int | None = None
    yellow_toner_status: int | None = None
    yellow_toner: int | None = None
